/*
 * MHSPI.cpp
 *
 *  Created on: Dec 12, 2017
 *      Author: captain
 */

/*
 * Copyright (c) 2010 by Cristian Maglie <c.maglie@bug.st>
 * SPI Master library for arduino.
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1, both as
 * published by the Free Software Foundation.
 */

#include <MHSPI.h>
//#include <MHUtility.h>

//#include "wiring_private.h"
#include "inc/hw_memmap.h"
#include "inc/hw_ssi.h"
#include "inc/hw_types.h"
#include "driverlib/ssi.h"
#include "driverlib/gpio.h"
#include "driverlib/sysctl.h"
#include "driverlib/pin_map.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"

#define SSIBASE g_ulSSIBase[SSIModule]
#define NOT_ACTIVE 0xA

static const unsigned long g_ulSSIBase[] = {
#if defined(__TM4C1294NCPDT__)
    SSI0_BASE, SSI1_BASE, SSI2_BASE, SSI3_BASE, SSI3_BASE
#endif
};

//*****************************************************************************
// The list of SSI peripherals.
//*****************************************************************************
static const unsigned long g_ulSSIPeriph[] = {
#if defined(__TM4C1294NCPDT__)
    SYSCTL_PERIPH_SSI0, SYSCTL_PERIPH_SSI1, SYSCTL_PERIPH_SSI2, SYSCTL_PERIPH_SSI3, SYSCTL_PERIPH_SSI3
#endif

};

//*****************************************************************************
// The list of SSI gpio configurations.
//*****************************************************************************
static const unsigned long g_ulSSIConfig[][4] = {
#if defined(__TM4C1294NCPDT__)
// from Table 17-1. SSI Signals (128TQFP)
    {GPIO_PA2_SSI0CLK, GPIO_PA3_SSI0FSS, GPIO_PA4_SSI0XDAT0, GPIO_PA5_SSI0XDAT1},
    {GPIO_PB5_SSI1CLK, GPIO_PB4_SSI1FSS, GPIO_PE4_SSI1XDAT0, GPIO_PE5_SSI1XDAT1},
    {GPIO_PD3_SSI2CLK, GPIO_PD2_SSI2FSS, GPIO_PD1_SSI2XDAT0, GPIO_PD0_SSI2XDAT1},
    {GPIO_PF3_SSI3CLK, GPIO_PF2_SSI3FSS, GPIO_PF1_SSI3XDAT0, GPIO_PF0_SSI3XDAT1},
    {GPIO_PQ0_SSI3CLK, GPIO_PQ1_SSI3FSS, GPIO_PQ2_SSI3XDAT0, GPIO_PQ3_SSI3XDAT1}
#endif
};

//*****************************************************************************
// The list of SSI gpio port bases.
//*****************************************************************************
static const unsigned long g_ulSSIPort[] = {
#if defined(__TM4C1294NCPDT__)
    GPIO_PORTA_BASE, GPIO_PORTB_BASE, GPIO_PORTD_BASE, GPIO_PORTF_BASE, GPIO_PORTQ_BASE
#endif
};

//*****************************************************************************
// The list of SSI gpio configurations.
//*****************************************************************************
static const unsigned long g_ulSSIPins[] = {
#if defined(__TM4C1294NCPDT__)
    GPIO_PIN_2 | GPIO_PIN_3 | GPIO_PIN_4 | GPIO_PIN_5,
    GPIO_PIN_5 | GPIO_PIN_4 | GPIO_PIN_4 | GPIO_PIN_5,
    GPIO_PIN_3 | GPIO_PIN_2 | GPIO_PIN_1 | GPIO_PIN_0,
    GPIO_PIN_3 | GPIO_PIN_2 | GPIO_PIN_1 | GPIO_PIN_0,
    GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_2 | GPIO_PIN_3
#endif
};

SPIClass::SPIClass(void) {
    SSIModule = NOT_ACTIVE;
    SSIBitOrder = MSBFIRST;
}

SPIClass::SPIClass(uint8_t module) {
    SSIModule = module;
    SSIBitOrder = MSBFIRST;
}

void SPIClass::begin() {
    unsigned long initialData = 0;
    uint32_t clockFreq = 0;
    uint32_t bitRate = 0;

    if(SSIModule == NOT_ACTIVE) {
        SSIModule = BOOST_PACK_SPI;
    }

    ROM_SysCtlPeripheralEnable(g_ulSSIPeriph[SSIModule]);
    ROM_SSIDisable(SSIBASE);
    ROM_GPIOPinConfigure(g_ulSSIConfig[SSIModule][0]);
    ROM_GPIOPinConfigure(g_ulSSIConfig[SSIModule][1]);
    ROM_GPIOPinConfigure(g_ulSSIConfig[SSIModule][2]);
    ROM_GPIOPinConfigure(g_ulSSIConfig[SSIModule][3]);
#if defined(__TM4C129XNCZAD__) || defined(__TM4C1294NCPDT__)
    if (SSIModule == 1) { // 1 is a split port
        ROM_GPIOPinTypeSSI(GPIO_PORTB_BASE, GPIO_PIN_5 | GPIO_PIN_4);
        ROM_GPIOPinTypeSSI(GPIO_PORTE_BASE, GPIO_PIN_4 | GPIO_PIN_5);
    } else {
        ROM_GPIOPinTypeSSI(g_ulSSIPort[SSIModule], g_ulSSIPins[SSIModule]);
    }
#endif


    /*
      Polarity Phase        Mode
         0     0   SSI_FRF_MOTO_MODE_0
         0     1   SSI_FRF_MOTO_MODE_1
         1     0   SSI_FRF_MOTO_MODE_2
         1     1   SSI_FRF_MOTO_MODE_3
    */

    /*
     * Default to
     * System Clock, SPI_MODE_0, MASTER,
     * 4MHz bit rate, and 8 bit data
    */
    // NOTE:
    // Sets the data clock source for the specified SSI peripheral
    // SSI_CLOCK_SYSTEM = system clock
    // SSI_CLOCK_PIOSC = precision internal oscillator
    ROM_SSIClockSourceSet(SSIBASE, SSI_CLOCK_SYSTEM);

    // Save current clock frequency
    //clockFreq = getClkFreq();
    clockFreq = 120000000L;

    // Set Bit Rate
    if (clockFreq > 4000000) // Only want a MAXIMUM BitRate of 4 MHz
        bitRate = 4000000;
    else
        bitRate = clockFreq/2;

    // Configures the synchronous serial interface
    // NOTE: THIS MUST BE TRUE: FSSI >= 2 * bit rate (master mode)
    ROM_SSIConfigSetExpClk(SSIBASE, clockFreq, SSI_FRF_MOTO_MODE_0, SSI_MODE_MASTER, bitRate, 8);

    ROM_SSIEnable(SSIBASE);

    //clear out any initial data that might be present in the RX FIFO
    while(ROM_SSIDataGetNonBlocking(SSIBASE, &initialData));
}

void SPIClass::end() {
    ROM_SSIDisable(SSIBASE);
}

void SPIClass::setBitOrder(uint8_t ssPin, uint8_t bitOrder) {
    SSIBitOrder = bitOrder;
}

void SPIClass::setBitOrder(uint8_t bitOrder) {
    SSIBitOrder = bitOrder;
}

void SPIClass::setDataMode(uint8_t mode) {
    HWREG(SSIBASE + SSI_O_CR0) &= ~(SSI_CR0_SPO | SSI_CR0_SPH);
    HWREG(SSIBASE + SSI_O_CR0) |= mode;
}

void SPIClass::setClockDivider(uint8_t divider){
  //value must be even
  HWREG(SSIBASE + SSI_O_CPSR) = divider;
}

uint8_t SPIClass::transfer(uint8_t data) {
    unsigned long rxtxData;

    rxtxData = data;
    if(SSIBitOrder == LSBFIRST) {
        asm("rbit %0, %1" : "=r" (rxtxData) : "r" (rxtxData));  // reverse order of 32 bits
        asm("rev %0, %1" : "=r" (rxtxData) : "r" (rxtxData));   // reverse order of bytes to get original bits into lowest byte
    }
    ROM_SSIDataPut(SSIBASE, (uint8_t) rxtxData);

    while(ROM_SSIBusy(SSIBASE));

    ROM_SSIDataGet(SSIBASE, &rxtxData);
    if(SSIBitOrder == LSBFIRST) {
        asm("rbit %0, %1" : "=r" (rxtxData) : "r" (rxtxData));  // reverse order of 32 bits
        asm("rev %0, %1" : "=r" (rxtxData) : "r" (rxtxData));   // reverse order of bytes to get original bits into lowest byte
    }

    return (uint8_t) rxtxData;
}

void SPIClass::setModule(uint8_t module) {
    SSIModule = module;
    begin();
}

SPIClass SPI;

